! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module cal


!----------------------------------------------------------------- 
! Routines handling the eROSITA CALDB
! Author:   GLamer, AIP           date: 06-JUL-2010
! Remarks:  PSF specific routines are compiled in module PSFLIB
! 26/10/2012 : add cal_scc2eroday, cal_eroday2scc
!
!
! subroutines:
! cal_distcorr
! cal_getboresight
! cal_i4toex
! cal_r4toex
! cal_stringtoex
! cal_filename
! tail_filename
! cal_setpar
! cal_getinstpar
! cal_sphere2tan
! cal_tan2sphere
! cal_getoffax
! cal_scc2eroday
!
!----------------------------------------------------------------- 


  use param
  use numkind
  use cfitsio_ifaces
  use fitsimage

  implicit none


  real(kind=real64),parameter,private       ::  deg2rad=1.74532925199D-02, rad2deg=57.29577951322  

  public 

  ! also defined in tel_io_mod, put into a separate module??
  real(kind=real64),parameter        ::  cal_deg2rad=1.74532925199D-02, cal_rad2deg=57.29577951322
  !
  real(kind=real64),parameter        ::  equpole_ra=180.D0, equpole_dec=90.D0




  type calhdr_type
    !sequence
    character(len=8)      :: filter
    real(kind=real32)     :: backgrnd
    real(kind=real32)     :: sumrcts
    real(kind=real32)     :: energy
    real(kind=real32)     :: theta
    real(kind=real32)     :: phi    
    real(kind=real32)     :: energ_lo
    real(kind=real32)     :: energ_hi
    integer(kind=int32)   :: chanmin
    integer(kind=int32)   :: chanmax
    character(len=8)      :: chantype
    character(len=8)      :: ccls0001
    character(len=8)      :: cdtp0001 
    character(len=16)     :: ccnm0001 
    character(len=68)     :: cbd10001  
    character(len=68)     :: cbd20001 
    character(len=68)     :: cbd30001    
    character(len=10)     :: cvsd0001    
    character(len=8)      :: cvst0001  
    character(len=68)     :: cdes0001 


    character(len=47) :: c_filter  
    character(len=47) :: c_backgrnd
    character(len=47) :: c_sumrcts
    character(len=47) :: c_energy
    character(len=47) :: c_theta
    character(len=47) :: c_phi
    character(len=47) :: c_energ_lo
    character(len=47) :: c_energ_hi
    character(len=47) :: c_chanmin 
    character(len=47) :: c_chanmax 
    character(len=47) :: c_chantype
    character(len=47) :: c_ccls0001
    character(len=47) :: c_cdtp0001
    character(len=47) :: c_ccnm0001
    character(len=47) :: c_cbd10001
    character(len=47) :: c_cbd20001
    character(len=47) :: c_cbd30001
    character(len=47) :: c_cvsd0001
    character(len=47) :: c_cvst0001
    character(len=47) :: c_cdes0001

    logical(kind=log32)   :: e_filter  
    logical(kind=log32)   :: e_backgrnd
    logical(kind=log32)   :: e_sumrcts
    logical(kind=log32)   :: e_energy
    logical(kind=log32)   :: e_theta
    logical(kind=log32)   :: e_phi
    logical(kind=log32)   :: e_energ_lo
    logical(kind=log32)   :: e_energ_hi
    logical(kind=log32)   :: e_chanmin 
    logical(kind=log32)   :: e_chanmax 
    logical(kind=log32)   :: e_chantype
    logical(kind=log32)   :: e_ccls0001
    logical(kind=log32)   :: e_cdtp0001
    logical(kind=log32)   :: e_ccnm0001
    logical(kind=log32)   :: e_cbd10001
    logical(kind=log32)   :: e_cbd20001
    logical(kind=log32)   :: e_cbd30001
    logical(kind=log32)   :: e_cvsd0001
    logical(kind=log32)   :: e_cvst0001
    logical(kind=log32)   :: e_cdes0001

  end type calhdr_type

!----------------------------------------------------------------------------------------------

  ! Instrument boresight, Euler angles phi, theta, psi:
  ! Obsolete, boresight now part of INSTPAR 
  type ibsrow_type
    !sequence
    character(len=16)   :: instrument_id
    real(kind=real64)   :: euphi
    real(kind=real64)   :: eutheta
    real(kind=real64)   :: eupsi
    real(kind=real64)   :: euphi_err
    real(kind=real64)   :: eutheta_err
    real(kind=real64)   :: eupsi_err
  end type ibsrow_type


  type boresight_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'BORESIGHT'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    type(calhdr_type)                :: calhdr
    integer(kind=int32)              :: ncol=7
    character(len=16), dimension(7)  :: cname = [ character(len=16) :: 'INSTRUMENT_ID', 'PHI', 'THETA', 'PSI', &
                                                                         'PHI_ERR', 'THETA_ERR', 'PSI_ERR']
    character(len=4) , dimension(7) :: cform = [ character(len=4)  :: '16A','1D','1D','1D','1D','1D','1D']
    character(len=16), dimension(7) :: cunit= [ character(len=16)  :: ' ','rad','rad','rad','rad','rad','rad' ]
    type(ibsrow_type), pointer, dimension(:)  :: data  => NULL()  
  end type boresight_type

!----------------------------------------------------------------------------------------------

  !---------------------------------------------------------- 
  ! Type definitions for star tracker boresight calibration files.  
  ! Boresight data are stored as quaternions.
  ! 
  ! 
  !-------------------------------------------------------------------------------------------- 

  type boresight_quatrow_type
 !   !sequence
    real(kind=real64)       ::  qx         !  quaternion x value
    real(kind=real64)       ::  qy         !  quaternion y value
    real(kind=real64)       ::  qz         !  quaternion z value
    real(kind=real64)       ::  qw         !  quaternion real part
    real(kind=real64)       ::  timelag    !  time lag events - attitude_source 
  end type boresight_quatrow_type

  type boresight_quat_type
!    !sequence
    character(len=clong)             :: filename    
    character(len=16)                :: extname = ' '
    type(shapehdr_type)              :: shapehdr 
    type(stdhdr_type)                :: stdhdr
    type(calhdr_type)                :: calhdr
    integer(kind=int32)              :: ncol=5      
    character(len=16), dimension(5)  :: cname = [ character(len=16) ::  'QX', 'QY' , 'QZ', 'QW', 'TIMELAG' ]
    character(len=4), dimension(5)   :: cform = [ character(len=4)  ::  '1D','1D','1D','1D','1D' ]
    character(len=16), dimension(5)  :: cunit = [ character(len=16) ::  ' ',' ',' ',' ','s' ]
    type(boresight_quatrow_type), pointer, dimension(:)  :: data  => NULL()
 !   type(boresight_quatrow_type), dimension(1)  :: data
  end type boresight_quat_type






 !----------------------------------------------------------------------------------------------- 


  type instpar_type
    !sequence
    real(kind=real64)   :: focal_length
    real(kind=real64)   :: mm_xpix    
    real(kind=real64)   :: mm_ypix
    real(kind=real64)   :: x_platescale
    real(kind=real64)   :: y_platescale
    integer(kind=int32) :: x_ccdpix
    integer(kind=int32) :: y_ccdpix
    real(kind=real64)   :: x_ref
    real(kind=real64)   :: y_ref
    real(kind=real64)   :: x_optax
    real(kind=real64)   :: y_optax
    real(kind=real32)   :: fov_rad 

   !----------------------------------------------------------------------
   !  Parameters for optical distortion
   !  dist_a00, dist_b00 : Symmetry axis of distortions
   !  (u,v) RAW CCD pixels -> (x,y) Corrected pixels coordinates
   !
   !  x =  u + SUM_p(dist_a_p0*(u - dist_a00)^p)
   !  y =  v + SUM_p(dist_b_q0*(u - dist_b00)^q)
 
   !  Reverse transformation
   !   u = x +  SUM_p(dist_ar_p0 * (x - dist_a00)^p) 
   !   v = y +  SUM_p(dist_br_p0 * (y - dist_a00)^p) 
   !   
   !  See also http://tdc-www.harvard.edu/wcstools/SIP_distortion_v1_0.pdf
   !----------------------------------------------------------------------

    real(kind=real64)   :: dist_a00
    real(kind=real64)   :: dist_a20
    real(kind=real64)   :: dist_a30
    real(kind=real64)   :: dist_b00 
    real(kind=real64)   :: dist_b02
    real(kind=real64)   :: dist_b03
    real(kind=real64)   :: dist_ar10
    real(kind=real64)   :: dist_ar20
    real(kind=real64)   :: dist_ar30
    real(kind=real64)   :: dist_br01
    real(kind=real64)   :: dist_br02
    real(kind=real64)   :: dist_br03
    real(kind=real64)   :: boresight_roll
    real(kind=real64)   :: boresight_x
    real(kind=real64)   :: boresight_y
    real(kind=real64)   :: timelag
  end type instpar_type


  type cal_par_type
    !sequence
    integer(kind=int32) :: chatter
    character(len=30)   :: codenam
    character(len=30)   :: telescop
    character(len=30)   :: instrume
    character(len=30)   :: detnam
    character(len=30)   :: filter
    character(len=10)   :: strtdate
    character(len=10)   :: strtime
    character(len=10)   :: stpdate
    character(len=10)   :: stptime
    character(len=256)  :: expression=''
    integer(kind=int32) :: maxret
    integer(kind=int32) :: nret
    integer(kind=int32) :: nfound   
    character(len=256), dimension(100)   :: history
    integer(kind=int32) :: nhist=0           
  end type cal_par_type

  contains



!--------------------------------------------------------------------------------

  subroutine cal_distcorr(instpar,x_in,y_in,x_out,y_out,ierr)

   !---------------------------------------------------------------
   ! Apply distortion correction to raw detector pixel coordinates    
   !  
   ! Distortion parameters are defined in CAL file INSTPAR
   ! 
   !  dist_a00, dist_b00 : Symmetry axis of distortions (usually optical axis)
   !  (u,v) RAW CCD pixels -> (x,y) Corrected pixels coordinates
   !
   !
   !  x =  u + SUM_p(dist_a_p0*(u - dist_a00)^p)
   !  y =  v + SUM_p(dist_b_q0*(u - dist_b00)^q)
 
   !  Reverse transformation
   !   u = x +  SUM_p(dist_ar_p0 * (x - dist_a00)^p) 
   !   v = y +  SUM_p(dist_br_p0 * (y - dist_a00)^p) 
   !   
   !  See also http://tdc-www.harvard.edu/wcstools/SIP_distortion_v1_0.pdf
   !  No cross-terms are used here
   !----------------------------------------------------------------------  


    implicit none
    type(instpar_type) , intent(in)            :: instpar 
!   real(kind=real64), pointer, dimension(:)   :: x_in,y_in
!   real(kind=real64), pointer, dimension(:)   :: x_out,y_out
    real(kind=real64)                          :: x_in,y_in
    real(kind=real64)                          :: x_out,y_out
    integer(kind=int32)                        :: direction
    integer(kind=int32)                        :: ierr
    real(kind=real64)                          :: dx,dy   
    character(len=*), parameter                :: rtname='CAL_DISTCORR'


    ierr=0

    if (instpar%dist_a20 == 0.D0 .and.  instpar%dist_a30 == 0.D0 .and.  &
        instpar%dist_b02 == 0.D0 .and.  instpar%dist_b03 == 0.D0 ) then 

       x_out = x_in
       y_out = y_in
    
    else

      dx =  x_in - instpar%dist_a00
      dy =  y_in - instpar%dist_b00 

      x_out = x_in + instpar%dist_a20*dx*dx + instpar%dist_a30*dx**3 
      y_out = y_in + instpar%dist_b02*dy*dy + instpar%dist_b03*dy**3 
 

    endif
      

  end subroutine cal_distcorr

!--------------------------------------------------------------------------------

  subroutine cal_getboresight(filename,ext,table,ierr)

    !-------------------------------------------------
    ! Obsolete (23-02-2015)
    ! -> boresight part of INSTPAR CAL files
    !-------------------------------------------------    

    implicit none


    type(boresight_type)             :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'CAL_GETBORESIGHT'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        call fread_stdhdr(unit,table%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVS(unit,colnum(1),i1,1,rowbuf,cnull,table%data(i1:i2)%instrument_id,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,table%data(i1:i2)%euphi,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,table%data(i1:i2)%eutheta,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,table%data(i1:i2)%eupsi,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(5) > 0) call  FTGCVD(unit,colnum(5),i1,1,rowbuf,r8null,table%data(i1:i2)%euphi_err,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(6) > 0) call  FTGCVD(unit,colnum(6),i1,1,rowbuf,r8null,table%data(i1:i2)%eutheta,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(7) > 0) call  FTGCVD(unit,colnum(7),i1,1,rowbuf,r8null,table%data(i1:i2)%eupsi_err,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    

  end   subroutine cal_getboresight


!--------------------------------------------------------------------------------



  subroutine cal_i4toex(parname,parval,expression,ierr)
  
  !--------------------------------------------------------------
  ! 
  ! Add numerical parameter to expression string for CALDB getcalf.
  ! The parameter is first converted from real32 to nearest integer. 
  ! NOTE: Resulting expression string must not be interrupted by blanks. 
  !       Make sure that 'expression' is initialized to '' or ' ' before
  !       first call.
  !            
  !--------------------------------------------------------------

    character(len=*)    :: parname
    real(kind=int32)    :: parval 
    character(len=*)    :: expression
    integer(kind=int32) :: ierr
    
    character(len=8)    :: cval
    character(len=len(parname)) :: chname
    character(len=*), parameter    :: rtname='CAL_I4TOEX'

    ierr=0
   
    write(cval,'(i8)') parval    

    cval=adjustl(cval) 
    chname=adjustl(parname)    

    if (len_trim(expression) == 0) then
      expression=trim(chname)//".eq."//trim(cval) 
    else
      expression=trim(expression)//".and."//trim(chname)//".eq."//trim(cval)
    endif 

  end  subroutine cal_i4toex



!--------------------------------------------------------------------------------


  subroutine cal_getboresight_quat(filename,iext,table,ierr)

    !-------------------------------------------------
    ! Read start tracker boresight CAL file 
    ! in quaternion format
    ! 
    !-------------------------------------------------    

    implicit none


    type(boresight_quat_type)        :: table
    character(len=*), intent(in)     :: filename
    integer(kind=int32), intent(in)  :: iext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=3)                 :: istr
    character(len=*),parameter       :: rtname = 'CAL_GETBORESIGHT_QUAT'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        call fread_stdhdr(unit,table%stdhdr,status)
        status=0
        call FTMAHD(unit,iext+1, hdutype,status)
       
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          write(istr,'(I3)') iext
          errstr = 'error moving to extension '//istr
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2

          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo

          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%qx,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,table%data(i1:i2)%qy,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,table%data(i1:i2)%qz,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,table%data(i1:i2)%qw,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(5) > 0) call  FTGCVD(unit,colnum(5),i1,1,rowbuf,r8null,table%data(i1:i2)%timelag,fnull,status) 
            fstat = max(fstat,status) 

            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    

  end subroutine cal_getboresight_quat

 !----------------------------------------------------------------------------------------------- 


  subroutine cal_r4toex(parname,parval,expression,ierr)
  
  !--------------------------------------------------------------
  ! 
  ! Add numerical parameter to expression string for CALDB getcalf.
  ! The parameter value is rounded with 2 fractional digits.
  ! NOTE: Resulting expression string must not be interrupted by blanks. 
  !       Make sure that 'expression' is initialized to '' or ' ' before
  !       first call.
  !            
  !--------------------------------------------------------------

    character(len=*)    :: parname
    real(kind=real32)   :: parval 
    character(len=*)    :: expression
    integer(kind=int32) :: ierr
    
    integer(kind=int32) :: ival
    character(len=12)    :: cval
    character(len=len(parname)) :: chname
    character(len=*), parameter :: rtname='CAL_R4TOEX'

    ierr=0
   
    write(cval,'(F12.2)') parval    

    cval=adjustl(cval) 
    chname=adjustl(parname)    

    if (len_trim(expression) == 0) then
      expression=trim(chname)//".eq."//trim(cval) 
    else
      expression=trim(expression)//".and."//trim(chname)//".eq."//trim(cval)
    endif 

  end  subroutine cal_r4toex

!--------------------------------------------------------------------------------


  subroutine cal_stringtoex(parname,parval,expression,ierr)   
  
  !--------------------------------------------------------------
  ! Add string parameter to expression string for CALDB getcalf.
  ! NOTE: Resulting expression string must not be interrupted by blanks 
  !       and string constants must be included in double quotation marks.
  !       Make sure that 'expression' is initialized to '' or ' ' before
  !       first call.
  ! parname:  Name of parmater (e.g. 'OBSMODE')
  ! parval :  string parameter value (e.g. 'POINTING') 
  ! expression: expression string as expected by gtcalf
  ! ierr  : error status     
  !--------------------------------------------------------------

    implicit none

    character(len=*)    :: parname
    character(len=*)    :: parval 
    character(len=*)    :: expression
    integer(kind=int32) :: ierr
    
    integer(kind=int32) :: ival
    character(len=len(parval))  :: cval
    character(len=len(parname)) :: chname
    character(len=*), parameter    :: rtname='CAL_STRINGTOEX'


    ierr=0
   
    cval=adjustl(parval) 
    chname=adjustl(parname)    

    if (len_trim(expression) == 0) then
      expression=trim(chname)//".eq."//"'"//trim(cval)//"'"
!      expression=trim(chname)//".eq."//trim(cval)
    else
      expression=trim(expression)//'.and.'//trim(chname)//'.eq.'//'"'//trim(cval)//'"'
!       expression=trim(expression)//'.and.'//trim(chname)//'.eq.'//trim(cval)
    endif 

  end  subroutine cal_stringtoex

!--------------------------------------------------------------------------------




  subroutine cal_filename(cpar,filename,ext,ierr)


  !---------------------------------------------------------------------------
  ! Determine CAL file name according to parameters in CAL_PAR and EXPRESSION
  ! Call CALDB routine GTCALF
  ! Return FILENAME, EXT  
  ! Store <FILENAME>+<EXT> in string CAL_PAR%HISTORY
  !
  ! NOTE: In case  GTCALF returns more than 1 file, currently the first file
  ! only is returned by  CAL_FILENAME.
  ! 
  !---------------------------------------------------------------------------

    implicit none

    type(cal_par_type), intent(inout)      :: cpar
    character(len=*), intent(out)          :: filename 
    integer(kind=int32), intent(out)       :: ext,ierr
    character(len=len(filename))           :: tname
    integer(kind=int32), parameter         :: maxret=100
    integer(kind=int32)                    :: nret,nfound,status 
    character(len=128), dimension(maxret)  :: filenames,online
    integer(kind=int32), dimension(maxret) :: extno
    character(len=3)                       :: cext
    character(len=256)                     :: outstr
    character(len=*), parameter            :: rtname='CAL_FILENAME'            
                    
    ierr=0
 



    cpar%chatter=1
    cpar%maxret=100
 


    call gtcalf (cpar%chatter, cpar%telescop, cpar%instrume, cpar%detnam, cpar%filter, &
                 cpar%codenam, cpar%strtdate, cpar%strtime, cpar%stpdate, cpar%stptime, &
                 cpar%expression, cpar%maxret, filenames, extno,  online, &
                 cpar%nret, cpar%nfound, status)          
 
   

    filename=filenames(1)
    ext=extno(1) 

   
    if (status > 0) then
      ierr=3
    else if  (cpar%nfound > 1) then
      ierr=1
      call wrerr(rtname,ierr,'Warning: More than 1 matching calibration file found')
      outstr="Warning: return file "//trim(filename)
      call wrerr(rtname,ierr,trim(outstr))
    endif
  

!   write(*,*) "cal_filename, nfound= ", cpar%nfound
!   write(*,*) "cal_filename, filename= ", filename

  !!   update CAL history:
  !!  (GL 08/03/2012: remove for the time being, since 
  !!   ERMLDET writes too many entries here:
   ! write(cext,'(I3)') ext
   ! cpar%nhist=cpar%nhist+1
   ! cpar%history(cpar%nhist)=trim(filename)//'+'//adjustl(cext)
    
    
     

  end subroutine cal_filename

!-------------------------------------------------------------------------------

  subroutine tail_filename(filename,tname,ierr)

    implicit none

    character(len=*), intent(in)     :: filename
    character(len=*), intent(out)    :: tname
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: ind1,ind2
    character(len=*), parameter      :: rtname='TAIL_FILENAME'

    ierr=0

    ind1=index(filename,'/',back=.true.)
    ind2=len_trim(filename)
    if (ind2 > ind1) then
      tname=filename(ind1+1:ind2)
    else
      ierr=3
      call wrerr(rtname,ierr,'Error: empty filename')
    endif

  end subroutine tail_filename
    
!--------------------------------------------------------------------------------
   

  subroutine cal_setpar(header,cal_par,ierr)

  !---------------------------------------------------------------------------
  !
  ! Copy strings from standard header (type stdhdr_type)
  ! to structure cal_par.
  ! Re-format start/stop time strings to CALDB format.
  ! Set CALDB parameter CHATTER from environment variable  SASS_VERBOSITY
  ! strings cal_par%detnam and cal_par%filter are currently set to 'NONE'
  !
  ! NOTE: (05/12/2014 GL) in the case of merged data from more than one 
  ! camera (INSTRUME='merged'), the string from  HEADER%INSTRUMX(1) is copied
  ! into PAR%INSTRUME.
  !---------------------------------------------------------------------------

    implicit none

    type(stdhdr_type) , intent(in)     :: header
    type(cal_par_type), intent(inout)  :: cal_par
    integer(kind=int32), intent(out)   :: ierr
    character(len=6)                   :: cenv
    character(len=*), parameter        :: rtname="CAL_SETPAR"    
 
    ierr=0
    
    cal_par%telescop  = header%telescop

    if (header%instrume == 'merged' .or. header%instrume == 'MERGED') then
      if (header%ninst > 0) then
        cal_par%instrume = header%instrumx(1)
      else
        ierr=3
        call wrerr(rtname,ierr,'HEADER keyword INSTRUME=MERGED, but NINST < 1')
      endif
    else
      cal_par%instrume   = header%instrume
    endif

    cal_par%detnam    = '-'
    cal_par%filter    = '-'  
    cal_par%strtdate = header%date_obs(1:10)  
    cal_par%stpdate  = header%date_end(1:10)
    cal_par%strtime  = header%date_obs(12:19)  
    cal_par%stptime  = header%date_end(12:19)

    
    ! SASS verbosity -> caldb verbosity:
    call get_environment_variable("SASS_VERBOSITY", cenv)
    cal_par%chatter = 2
    if (cenv == "sparse" .or. cenv == "SPARSE")  cal_par%chatter = 1
    if (cenv == "chatty" .or. cenv == "CHATTY" )  cal_par%chatter = 5     
      
 

  end subroutine cal_setpar 

!--------------------------------------------------------------------------------


  subroutine cal_getinstpar(filename,ext,instpar,ierr)

  !-------------------------------------------------------------
  !
  ! Author: GLamer           date: 06-JUL-2010
  ! Read instrument specific telescope/detector geometry files
  ! Calibration table has one row per column
  ! Columns:
  ! FOCAL_LENGTH   :  telescope focal lengths in mm
  ! MM_XPIX        :  raw X CCD pixel size in mm
  ! MM_YPIX        :  raw Y CCD pixel size in mm 
  ! X_PLATESCALE   :  X raw pixel size in arcsec
  ! Y_PLATESCALE   :  Y raw pixel size in arcsec
  ! X_CCDPIX       :  Number of CCD pixels in X 
  ! Y_CCDPIX       :  Number of CCD pixels in Y
  ! X_REF          :  reference x pixel for telescope specific attitude 
  ! Y_REF          :  reference y pixel for telescope specific attitude 
  ! X_OPTAX        :  RAW X coordinate of optical axis intersection
  ! Y_OPTAX        :  RAW Y coordinate of optical axis intersection 
  !
  ! 
  ! 
  !---------------------------------------------------------------- 

    implicit none  

    character(len=*), intent(in)     :: filename
    integer(kind=int32)              :: ext
    type(instpar_type), intent(out)  :: instpar 
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: readwrite=0
    integer(kind=int32)              :: status,unit,blocksize,hdutype
    integer(kind=int32)              :: colnum,nrow,col
    logical(kind=log32)              :: anyf,casesens=.false.
    integer(kind=int32), dimension(1):: ival
    real(kind=real64), dimension(1)  :: dval
    real(kind=real32), dimension(1)  :: rval
    integer(kind=int32)              :: inull=0
    real(kind=real64)                :: dnull=0.D0
    real(kind=real32)                :: rnull=0.E0
    character(len=11)                :: rtname="CAL_GETINSTPAR"    
    
    

    ierr=0
    status=0  



    call ftgiou(unit,status)
    call ftopen(unit,filename,readwrite,blocksize,status)
    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error: could not open file')
    else
      ! For FTMAHD, the primary extension is ext=1, therefore add +1 
      call ftmahd(unit,ext+1,hdutype,status)
      if (status /= 0) then
        ierr=3
        call wrerr(rtname,ierr,'Error: FITS extension not found')
      else
        call ftgnrw(unit,nrow,status)
        if (nrow /= 1) then
          ierr=1
          call wrerr(rtname,ierr,'Warning: expected nrow=1')
        endif


        call ftgcno(unit, casesens,"FOCAL_LENGTH", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%focal_length = dval(1)

        call ftgcno(unit, casesens,"MM_XPIX", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%mm_xpix = dval(1)

        call ftgcno(unit, casesens,"MM_YPIX", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%mm_ypix = dval(1)

        call ftgcno(unit, casesens,"X_PLATESCALE", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%x_platescale = dval(1)

        call ftgcno(unit, casesens,"Y_PLATESCALE", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%y_platescale = dval(1)

        call ftgcno(unit, casesens,"X_CCDPIX", col, status)
        call ftgcvj(unit,col,1,1,1,inull,ival,anyf,status)
        instpar%x_ccdpix = ival(1)

        call ftgcno(unit, casesens,"Y_CCDPIX", col, status)
        call ftgcvj(unit,col,1,1,1,inull,ival,anyf,status)
        instpar%y_ccdpix = ival(1)


        call ftgcno(unit, casesens,"X_REF", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%x_ref = dval(1)

        call ftgcno(unit, casesens,"Y_REF", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%y_ref = dval(1)


        call ftgcno(unit, casesens,"X_OPTAX", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%x_optax = dval(1)

        call ftgcno(unit, casesens,"Y_OPTAX", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%y_optax = dval(1)

        call ftgcno(unit, casesens,"FOV_RAD", col, status)
        call ftgcve(unit,col,1,1,1,rnull,rval,anyf,status)
        instpar%fov_rad = rval(1)


        call ftgcno(unit, casesens,"DIST_A20", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_a20 = dval(1)

        call ftgcno(unit, casesens,"DIST_A30", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_a30 = dval(1)

        call ftgcno(unit, casesens,"DIST_B02", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_b02 = dval(1)

        call ftgcno(unit, casesens,"DIST_B03", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_b03 = dval(1)

        call ftgcno(unit, casesens,"DIST_AR10", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_ar10 = dval(1)

        call ftgcno(unit, casesens,"DIST_AR20", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_ar20 = dval(1)

        call ftgcno(unit, casesens,"DIST_AR30", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_ar30 = dval(1)

        call ftgcno(unit, casesens,"DIST_BR01", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_br01 = dval(1)

        call ftgcno(unit, casesens,"DIST_BR02", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_br02 = dval(1)

        call ftgcno(unit, casesens,"DIST_BR03", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%dist_br03 = dval(1)


        call ftgcno(unit, casesens,"BORESIGHT_ROLL", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%boresight_roll = dval(1)


        call ftgcno(unit, casesens,"BORESIGHT_X", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%boresight_x = dval(1)


        call ftgcno(unit, casesens,"BORESIGHT_y", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%boresight_y = dval(1)


        call ftgcno(unit, casesens,"TIMELAG", col, status)
        call ftgcvd(unit,col,1,1,1,dnull,dval,anyf,status)
        instpar%timelag = dval(1)



        if (status /= 0) then
          ierr=3
           call wrerr(rtname,ierr,'Error: table element(s) could not be read')
        endif
      endif 
    endif

  end  subroutine cal_getinstpar

!--------------------------------------------------------------------------------


  subroutine cal_xyrotate(x_in,y_in,x0,y0,phi,x_out,y_out,ierr)


    !----------------------------------------------------------
    ! Rotates position in cartesian coordinates by angle phi
    ! around  (x0,y0) counter-clockwise.
    ! Clockwise rotation is achieved by negative values of phi.
    ! Rotation angle phi is given in degrees.
    ! Equal (pixel) scaling in x,y is assumed.
    !----------------------------------------------------------    

    implicit none

    real(kind=real64), intent(in)   :: x_in,y_in,x0,y0,phi
    real(kind=real64), intent(out)  :: x_out,y_out 
    integer(kind=int32),intent(out) :: ierr
    
    real(kind=real64)               :: rsin,rcos,dx,dy             
    
    character(len=*), parameter     :: rtname='CAL_XYROTATE'

   
    ierr=0

    dx=x_in-x0
    dy=y_in-y0
    rsin = sin(phi*deg2rad)   
    rcos = cos(phi*deg2rad)   
        
    x_out = x0 +  dx*rcos - dy*rsin
    y_out = y0 +  dx*rsin + dy*rcos


  end  subroutine cal_xyrotate



!-------------------------------------------------------------------------------

  subroutine cal_sphere2tan(ra,dec,ra_ref,dec_ref,dx,dy,ierr)

!--------------------------------------------------------------------------------
! calculates tangential-plane approximation to spherical coordinates
! to transform delta-alpha, delta-delta into delta-x, delta-y
! from EXSAS/XMM-SAS routine sphere2
! 
! F90 version: GL 02/07/2010
!
! all RA, DEC values must be given in units of radians
! Input parameters:
!   RA         real64  source RA  [radian]
!   DEC        real64  source DEC [radian]
!   RA_REF     real64  reference RA [radian]
!   DEC_REF    real64  reference DEC [radian]
! Output parameters:
!  DX:         real64  angular separation in RA direction  (positive = East)
!  DY: angular real6   separation in DEC direction         (positive = North)
!  IERR:       int32   SASS error status
!--------------------------------------------------------------------------------

    implicit none

    real(kind=real64), intent(in)  :: ra,dec,ra_ref,dec_ref
    real(kind=real64), intent(out) :: dx,dy
    integer(kind=int32)            :: ierr

    real(kind=real64) ::  dcosde, dcosra, dh, diffra
    real(kind=real64) ::  dnsde, dncde, dsinde, dsinra
    real(kind=real64) ::  d1c, d1s, d2c, d2s
    character(len=*), parameter    :: rtname='CAL_SPHERE2TAN'


    ierr=0   

    diffra = ra - ra_ref
    dsinra = SIN(diffra)
    dcosra = COS(diffra)
    dsinde = SIN(dec)
    dcosde = COS(dec)
    dnsde =  SIN(dec_ref)
    dncde =  COS(dec_ref)

    d1s    =   dsinra * dcosde
    dh     = - dcosde * dcosra
    d1c    =   dnsde * dsinde - dh * dncde
    dx     =   ATAN2(d1s,d1c)     

    d2s    = dh * dnsde + dncde * dsinde
    d2c    = dsqrt(1.0d0 - d2s * d2s)
    dy = ATAN2(d2s,d2c)
 

  end subroutine cal_sphere2tan



  subroutine cal_tan2sphere(dx,dy,ra_ref,dec_ref,ra,dec,ierr)

!--------------------------------------------------------------------------------
! calculates spherical sky coordinates from x,y values in tangential projection 
! see Greisen: AIPS Memo 27 (1994)
! 
! dx , dy  values must be given in units of radians
! orientation dy -> positive dec (north), dx -> positive ra (east)
! Input parameters:
!  DX:         real64  angular separation in RA direction  (positive = East)
!  DY: angular real6   separation in DEC direction         (positive = North)
!   RA         real64  source RA  [radian]
!   DEC        real64  source DEC [radian]
!   RA_REF     real64  reference RA [radian]
!   DEC_REF    real64  reference DEC [radian]
! Output parameters:
!   RA         real64  source RA  [radian]
!   DEC        real64  source DEC [radian]
!  IERR:       int32   SASS error status
!--------------------------------------------------------------------------------

    implicit none

    real(kind=real64), intent(in)  :: dx,dy,ra_ref,dec_ref
    real(kind=real64), intent(out) :: ra,dec
    integer(kind=int32)            :: ierr

    real(kind=real64)              :: cosdec0,sindec0,dd
    real(kind=real64),parameter    :: dpi05=1.5707963267948966D0,dtwopi= 6.2831853071795862D0
    character(len=*), parameter    :: rtname='CAL_TAN2SPHERE'

    ierr=0   


    cosdec0=cos(dec_ref)
    sindec0=sin(dec_ref)


    dd=cosdec0-dy*sindec0   

    if (dd == 0) then
      ! poles of the spherical coordinate system:
      ra  =  0.D0
      dec =  sign(dpi05,dec_ref)  
    else    
      ra  = ra_ref + atan2(dx,dd)
      dec = atan(cos(ra - ra_ref)*(dy*cosdec0+sindec0)/dd)
    endif


  end subroutine cal_tan2sphere

!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------

  subroutine cal_angdist(ra1,ra2,dec1,dec2,dist,ierr)


  !--------------------------------------------------------------------------------
  ! calculates angular great circle distance between 2 equatorial coordinates
  ! ra1, dec1, ra2, dec2.
  ! Unit of all angles is radian
  ! Note: Limited accuracy for very small distances (~2E-6 for dist=1.0 arcsec)
  !
  !--------------------------------------------------------------------------------

    implicit none

    real(kind=real64)                :: ra1,ra2,dec1,dec2    
    real(kind=real64)                :: dist
    integer(kind=int32)              :: ierr
    real(kind=real64)                :: ddec,dra
    real(kind=real64),parameter      :: eps=1.D-8

    character(len=*), parameter    :: rtname='CAL_ANGDIST'    


    ierr=0

    if (abs(dec1) > 90.D0*deg2rad .or.  abs(dec2) > 90.D0*deg2rad) then
      ierr=4
      call wrerr(rtname,ierr,'dec1 or dec2 out of range')
    else
      ddec=dec2-dec1
      dra=ra2-ra1
      if ((ddec == 0.D0) .and. (dra == 0.D0)) then 
        dist=0.D0
      elseif (abs(ddec) < eps .and. abs(ddec) < eps) then
        dist=(ddec*ddec+cos(dec1)*dra*dra)**0.5
      else
        dist=acos(sin(dec1)*sin(dec2)+cos(dec1)*cos(dec2)*cos(ra2-ra1))
      endif
    endif
    
    


  end subroutine cal_angdist

!-------------------------------------------------------------------------------




  subroutine cal_getoffax(ra,dec,header,theta,phi,ierr)

  !--------------------------------------------------------------------------------
  ! Calculates off-axis angle theta and position angle phi of a source from 
  ! source RA, DEC, and the RA_PNT, DEC_PNT values in the image header.
  ! For a pointed observation the standard eROSITA header contains the mean 
  ! direction of the optical axis for each telescope in the keywords RA_PNT, DEC_PNT.
  ! Input parameters:
  !   RA:     real64               RA [degrees]
  !   DEC:    real64               DEC[degrees]
  !   HEADER: type(stdhdr_type)    standard eROSITA header 
  !   IERR  : int32                SASS error status
  ! Output parameters:
  !   THETA: real32                source off-axis angle in arcmin
  !   PHI:   real32                source position angle in deg
  !                                (counter-clockwise from North)  
  !   
  !  Change phi, theta to real32  17/04/2012  (GL)
  !---------------------------------------------------------------------------------        

    implicit none
 
    real(kind=real64),  intent(in)     :: ra, dec   
    type(stdhdr_type),  intent(in)     :: header
    real(kind=real32),  intent(out)    :: theta, phi
    integer(kind=int32),intent(out)    :: ierr
    real(kind=real64), parameter       :: deg2rad =  0.0174532925199433   
    real(kind=real64)                  :: rra, rdec, rra_pnt,rdec_pnt
    real(kind=real64)                  :: dx,dy,dra,ddec
    integer(kind=int32)                :: status
    character(len=*), parameter        :: rtname='CAL_GETOFFAX'

    ierr = 0
     
    rra=ra*deg2rad
    rdec=dec*deg2rad
    rra_pnt =header%ra_pnt*deg2rad
    rdec_pnt=header%dec_pnt*deg2rad
  
    call cal_sphere2tan(rra,rdec,rra_pnt,rdec_pnt,dx,dy,status)

    ierr=status
    
    
    theta = SQRT(dx*dx+dy*dy)/deg2rad*60.D0 
      
    phi=atan2(-dx,dy)/deg2rad


    ierr=status

        
    
  end subroutine cal_getoffax


!-------------------------------------------------------------------------------

  subroutine cal_scc2eroday(t_scc,eroday,ierr)

    implicit none

    real(kind=real64)                        :: t_scc
    integer(kind=int32)                      :: eroday,ierr
    ! define erodays here, read from cal file in future ?
    ! 1 ero_day = 4 hrs?
    real(kind=real64),parameter              :: t_day0=0.D0,eroday_sec=14400.D0
  
    character(len=*), parameter              :: rtname='CAL_SCC2ERODAY'  
    

    ierr=0
 
    eroday = int((t_scc-t_day0)/eroday_sec)
    

  end subroutine  cal_scc2eroday


!-------------------------------------------------------------------------------

  subroutine cal_eroday2scc(eroday,tstart,tend,ierr)

    implicit none

    integer(kind=int32)                      :: eroday
    real(kind=real64)                        :: tstart,tend
    integer(kind=int32)                      :: ierr
    ! define erodays here, read from cal file in future ?
    ! 1 ero_day = 4 hrs?
    real(kind=real64),parameter              :: t_day0=0.D0,eroday_sec=14400.D0
    character(len=*), parameter              :: rtname='CAL_ERODAY2SCC' 

    ierr=0
   
    tstart=real(t_day0)+real(eroday)*eroday_sec
    tend=real(t_day0)+real(eroday+1)*eroday_sec 

    

  end subroutine  cal_eroday2scc

!-------------------------------------------------------------------------------   




  subroutine fwrite_calhdr(unit,header,ierr)
 

    type(calhdr_type)             :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    integer(kind=int32)               :: i
    character(len=1)                  :: strnum
    character(len=8)                  :: keyname  
    character(len=64) :: version    
    character(len=80)                 :: taskname
!    integer(kind=int32),parameter     :: decimals = 15
    character(len=*), parameter       :: rtname = 'fwrite_calhdr'


    ierr=0
    status=0

    call FTUKYS(unit,'FILTER',header%filter                ,header%c_filter ,status)
    call FTUKYE(unit,'backgrnd',header%backgrnd, decimals  ,header%c_backgrnd,status)
    call FTUKYE(unit,'sumrcts',header%sumrcts,   decimals  ,header%c_sumrcts,status)
    call FTUKYE(unit,'energy',header%energy,     decimals  ,header%c_energy,status)
    call FTUKYE(unit,'theta',header%theta,       decimals  ,header%c_theta,status)
    call FTUKYE(unit,'phi',header%phi,           decimals  ,header%c_phi,status)
    call FTUKYE(unit,'energ_lo',header%energ_lo, decimals  ,header%c_energ_lo,status)
    call FTUKYE(unit,'energ_hi',header%energ_hi, decimals  ,header%c_energ_hi,status)
    call FTUKYJ(unit,'chanmin',header%chanmin              ,header%c_chanmin,status)
    call FTUKYJ(unit,'chanmax',header%chanmax              ,header%c_chanmax,status)
    call FTUKYS(unit,'chantype',header%chantype            ,header%c_chantype ,status)
    call FTUKYS(unit,'ccls0001',header%ccls0001            ,header%c_ccls0001 ,status)
    call FTUKYS(unit,'cdtp0001',header%cdtp0001            ,header%c_cdtp0001 ,status)
    call FTUKYS(unit,'ccnm0001',header%ccnm0001            ,header%c_ccnm0001 ,status)
    call FTUKYS(unit,'cbd10001',header%cbd10001            ,header%c_cbd10001 ,status)
    call FTUKYS(unit,'cbd20001',header%cbd20001            ,header%c_cbd20001 ,status)
    call FTUKYS(unit,'cbd30001',header%cbd30001            ,header%c_cbd30001 ,status)
    call FTUKYS(unit,'cvsd0001',header%cvsd0001            ,header%c_cvsd0001 ,status)
    call FTUKYS(unit,'cvst0001',header%cvst0001            ,header%c_cvst0001 ,status)
    call FTUKYS(unit,'cdes0001',header%cdes0001            ,header%c_cdes0001 ,status)

    if (status > 0) ierr=3    

  end subroutine fwrite_calhdr


!-------------------------------------------------------------------------------



  subroutine fput_boresight(filename,append,table,ierr) 
 
  !------------------------------------------------------------
  ! Write boresight calibration file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(boresight_quat_type)          :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: nhdu, varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=1
    character(len=16), dimension(ncol) :: ttype,tunit
    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_BORESIGHT'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  

 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0

      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)
 
      call fwrite_calhdr(unit,table%calhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%qx,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%qy,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%qz,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%qw,status)
          fstat = max(fstat,status)
        enddo
      endif

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  
  end   subroutine fput_boresight
 

!-------------------------------------------------------------------------------






end module cal
